# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from airbyte_cdk.sources.declarative.requesters.query_properties.property_selector.json_schema_property_selector import (
    JsonSchemaPropertySelector,
)
from airbyte_cdk.sources.declarative.requesters.query_properties.property_selector.property_selector import (
    PropertySelector,
)

__all__ = ["JsonSchemaPropertySelector", "PropertySelector"]
